# Code coverage analysis on Unit Tests.
#
#
# Copyright (c) 2006, Ubiquitous Business Technology (http://ubit.com)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    * Neither the name of Ubit nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# == Author
# Zev Blut, Pierre Baumard

require "kwala"
require "#{ACTION_DIR}/loc_count"
require 'tempfile'

# run with
# ./kwala -p TestProj -d ~/projects/TestProj
# -a code_coverage -o ~/tmp/builder

class CodeCoverageAction < BuildAction
  include LOCCountModule

  CodeCoverageFileData = Struct.new(:ruby_file, :percent, :rcov_index,
                                    :file_cov_path, :loc)

  RCOV = '/usr/bin/env rcov'

  SUMMARY_COUNT = 30

  PERCENT_ERROR = 70

  PERCENT_WARNING = 90

  DEBUG = false

  def initialize
    @output_data = Array.new
  end

  def build_action(context)
    @ouput_dir = File.expand_path(context.output_directory)
    if !File.exists?("#{@ouput_dir}/coverage")
      FileUtils.mkdir("#{@ouput_dir}/coverage")
    end

    @test_cases_count = context.test_files.size

    run_test_cases(context)

    context.ruby_files.each do |ruby_file|
      test_path = test_index_path(context, ruby_file)
      rcov_index = test_path + "index.html"
      file_cov_path = test_path + file_coverage_filename(context, ruby_file)
      loc = count_loc(ruby_file).first
      if File.exist?(test_path)
        percent = parse_coverage_file(file_cov_path)
        @output_data << CodeCoverageFileData.new(ruby_file, percent,
                                                 rcov_index, file_cov_path, loc)
      else
        @output_data << CodeCoverageFileData.new(ruby_file, nil, rcov_index,
                                                 file_cov_path, loc)
      end
    end

  end


  def summary_display(context)
    template = TemplateFile.new(self.class.summary_template_file)
    unit_base = "#{context.project_name}_code_coverage.html"

    cov_loc_weighted_sum = @output_data.inject(0) do |sum, file_data|
      sum + (file_data.percent.to_i * file_data.loc)
    end
    loc_sum = @output_data.inject(0){|sum, file_data| sum + file_data.loc}

    global_cov = cov_loc_weighted_sum / loc_sum

    context.amrita_data[:code_coverage_results] = {
      :c_num => @test_cases_count,
      :global_cov => global_cov
    }

    output_summary = @output_data.sort do |x, y|
      non_covered_loc(y) <=> non_covered_loc(x)
    end
    context.amrita_data[:code_coverage_summary] = {
      :entry =>  entries(output_summary[0..SUMMARY_COUNT])
    }

    context.amrita_data[:code_coverage_details] =
     (Amrita::e(:a, :href => unit_base) { "Code Coverage Details" })
    summary_expand(template, context)
  end

  def detailed_display(context)
    template = TemplateFile.new(self.class.detailed_template_file)

    context.amrita_data[:code_coverage_details] = {
      :entry =>  entries(@output_data)
    }

    det_res = ProjectBuilderUtils.expand_template(template, context.amrita_data)
    det_file = "#{context.output_directory}/#{context.project_name}_code_coverage.html"
    [det_file, det_res]
  end

  ###########################
  # Protected Methods
  ###########################
  protected

  def non_covered_loc(file_data)
   (((100.0 - file_data.percent.to_f) / 100.0) * file_data.loc).to_i
  end

  def url_alias(url)
    url.sub(@ouput_dir, '').sub(/^(\/)*/, '')
  end

  def entries(output_data)
    output_data.map do |file|
      {
        :file => amrita_file(file.percent, file.ruby_file, file.rcov_index),
        :loc => file.loc.to_s,
        :nloc => non_covered_loc(file).to_s,
        :result =>  put_class(file.percent, amrita_result(file.percent, file.file_cov_path))
      }
    end

  end

  def put_class(percent, amrita_e)
    percent = percent.to_i
    if percent < PERCENT_ERROR
      Amrita.a(:class => "error") {amrita_e}
    elsif percent < PERCENT_WARNING
      Amrita.a(:class => "warning") {amrita_e}
    else
      Amrita.a(:class => "ok") {amrita_e}
    end
  end

  def amrita_file(percent, ruby_file, rcov_index)
    if percent
      Amrita.e(:a, :href => url_alias(rcov_index)) {ruby_file}
    else
      Amrita.e(:span, :title => "no file #{rcov_index}") {ruby_file}
    end
  end

  def amrita_result(percent, file_cov_path)
    if percent.to_i > 0
      Amrita.e(:a, :href => url_alias(file_cov_path)) {"#{percent} %"}
    else
      Amrita.e(:span, :title => "bad file #{file_cov_path}") {"0 %"}
    end
  end

  def test_index_path(context, file)
    proj_dir = context.project_directory.sub(/\/$/, '')
    test_file =  File.expand_path(proj_dir) + '_tests' +
    File.dirname(file).sub(proj_dir, '') +
    '_tc_' + File.basename(file)
    out_file_dir(test_file, context.output_directory)[1]
  end

  def file_coverage_filename(context, file)
    # from /usr/local/bin/rcov:74:in `normalize_filename' (rcov 0.5)
    file = File.expand_path(file).gsub(/^#{Regexp.escape(Dir.getwd)}\//, '')

    file.split("/").join("-").gsub('.', '_') + ".html"
  end

  COV_PATTERN = "<td class='covered' width='(\\d*)'"
  REGEXP = /#{COV_PATTERN}.*#{COV_PATTERN}/m

  def parse_coverage_file(file)
    if File.exists?(file) and (match = REGEXP.match(File.read(file)))
      match[2].to_i
    else
      0
    end
  end

  def out_file_dir(test_file, out_dir)
    f_out = test_file.split("/").join("_") + "_coverage"
    o_dir = "#{out_dir}/coverage/#{f_out}/"
    [f_out, o_dir]
  end

  def exclude_cmd(context)
    #if context.project_name =~ LIB_PROJECTS
    #  # discard the default regexp list, do not exclude site_ruby lib
    #  "--exclude-only lib/ruby/1.8"
    #else
      "--exclude lib/ruby"
    #end
  end

  def run_test_cases(context)
    context.test_files.each do |test_file|
      run_test_case_with_rcov(test_file, context)
    end
  end

  def run_test_case_with_rcov(test_file, context)
    test_file = File.expand_path(test_file)
    f_out, o_dir = out_file_dir(test_file, @ouput_dir)
    o_dir = File.expand_path(o_dir)

    if File.exists?(o_dir)
      FileUtils.rm_rf(o_dir)
    end
    FileUtils.mkdir(o_dir)

    cmd = "#{RCOV} #{exclude_cmd(context)} -o #{o_dir} #{test_file}"
    puts cmd if DEBUG
    system(cmd)
  end

end


# use --aggregate option of rcov >= 0.7
class AggregateCodeCoverageAction < CodeCoverageAction

  def self.summary_template_file
    superclass.summary_template_file
  end

  def self.detailed_template_file
    superclass.detailed_template_file
  end

  ##################
  # protected
  ##################
  protected

  def out_file_dir(test_file, out_dir)
    f_out = test_file.split("/").join("_") + "_coverage"
    [f_out, "#{out_dir}/coverage/aggregate/"]
  end

  def run_test_cases(context)
    test_files = context.test_files.dup.map {|file| File.expand_path(file)}

    o_dir = File.expand_path("#{File.expand_path(context.output_directory)}/coverage/aggregate")
    puts o_dir
    if File.exists?(o_dir)
      FileUtils.rm_rf(o_dir)
    end
    FileUtils.mkdir(o_dir)

    rcov_data = Tempfile.new('rcov_aggregate').path

    one_file = test_files.shift

    # first do not generate html, just aggregate to rcov file data
    test_files.each do |file|
      # "No file to analyze was found" with just --no-html
      cmd = "#{RCOV} #{exclude_cmd(context)} --no-html -t --aggregate #{rcov_data} #{file}"
      puts cmd if DEBUG
      system(cmd)
    end

    # generate HTML for last file to be run
    cmd = "#{RCOV} #{exclude_cmd(context)} --aggregate #{rcov_data} -o #{o_dir} #{one_file}"
    system(cmd)

  end

end

