# $Id: code_duplication.rb 29 2008-04-02 07:32:44Z zev $
#
# This file implements the logic to check for source code duplication.
#
# Requires the Java program PMD which uses the CPD tool.

# Copyright (c) 2006, Ubiquitous Business Technology (http://ubit.com)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    * Neither the name of Ubit nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# == Author
# Zev Blut (zb@ubit.com)


class CodeDuplicationAction < BuildAction

  def initialize
    @duplicates = nil
    @file_count = 0
  end

  def build_action(context)
    temporarily_remove_symlinks(context.project_directory) do |dir|
      @duplicates = find_duplicate_code(dir)
    end
    # Store file count for scoring
    @file_count = context.ruby_files.size + context.test_files.size
  end

  def summary_display(context)
    template = TemplateFile.new(self.class.summary_template_file)
    context.amrita_data[:duplicate_code_results] = {
      :duplicate_count => @duplicates.size,
      :duplicates => @duplicates
    }
    context.amrita_data[:code_duplication_details] =
     (Amrita::e(:a, :href=>"#{context.project_name}_code_duplication.html") { "Code Duplication Details" })
    summary_expand(template, context)
  end

  def detailed_display(context)
    det_res = expand_code_duplication_template(context.amrita_data[:duplicate_code_results])
    det_file = "#{context.output_directory}/#{context.project_name}_code_duplication.html"
    [det_file, det_res]
  end

  def score
    # This should be based on code size ratio, for now using file count, but
    # probably lines of code is better.
    if @duplicates.size > @file_count
      1
    else
      score_scaling(@duplicates.size, @file_count)
    end
  end

  def find_duplicate_code(dir=Dir.pwd)
    require "rexml/document"
    results = Array.new
    Open3.popen3( cpd_command(dir) ) do |stdin, stdout, stderr|
      xml = time_out_readlines(stdout)
      if cpd_error?(stderr) || xml.first == "IO readlines timed out."
        return results
      end
      doc = REXML::Document.new(xml.join("\n"))
      doc.elements.each("pmd-cpd/duplication") do |dup|
        results << {
          :duplicate =>  {
            :entry_num => results.size,
            :lines => dup.attribute("lines").to_s,
            :tokens => dup.attribute("tokens").to_s,
            :code_fragement => highlight_code(dup.elements["codefragment"].cdatas.to_s),
            :files => dup.get_elements("file").map { |f|
              {
                :path => f.attribute("path").to_s,
                :line => f.attribute("line").to_s
              }
            },
          }
        }
      end
    end
    results
  end


  def expand_code_duplication_template(data)
    template = TemplateFile.new(self.class.detailed_template_file)
    data[:duplicate_refs] = data[:duplicates].map do |d|
      {
        :entry_num => Amrita::a(:href=>"\##{d[:duplicate][:entry_num]}") {d[:duplicate][:entry_num]}
      }
    end
    max_size = (data[:duplicates].size - 1)
    data[:duplicates].each_with_index do |d, idx|
      d[:duplicate][:entry_num] = Amrita::a(:name=>"#{d[:duplicate][:entry_num]}") {d[:duplicate][:entry_num]}
      if idx > 0
        d[:duplicate][:prev_entry_num] = Amrita::a(:href=>"\##{idx - 1}") { "Prev" }
      end
      if idx < max_size
        d[:duplicate][:next_entry_num] = Amrita::a(:href=>"\##{idx + 1}") { "Next" }
      end
    end
    ProjectBuilderUtils.expand_template(template, data)
  end

  def highlight_code(code)
    # Check if enscript is installed.
    cmd = "/usr/bin/enscript --color --language=html -Eruby --output=-"
    out_data = nil
    Open3.popen3(cmd) do |stdin, stdout, stderr|
      stdin.puts code
      stdin.close
      out_data = time_out_readlines(stdout).join("")
      #time_out_readlines(stderr)
    end
    if (out_data && ( m = /(<pre>.*<\/pre>)/im.match(out_data) ) )
      code = Amrita::SanitizedString.new(m[1])
    end

    code
  end

  def cpd_command(dir)
    "java -Xmx256m -cp #{cpd_jar} net.sourceforge.pmd.cpd.CPD" +
      " --minimum-tokens #{minimum_tokens} --files #{dir}" +
      " --language ruby --format net.sourceforge.pmd.cpd.XMLRenderer"
  end

  def cpd_jar
    EXTERNAL_DIR + "pmd.jar"
  end

  def minimum_tokens
    50
  end

  def cpd_error?(stderr)
    lines = stderr.readlines
    if !lines.empty?
      STDERR.puts "code_duplication error with cpd:", lines
      true
    else
      false
    end
  end

  def temporarily_remove_symlinks(dir)
    symlinks = Hash.new

    Find.find(dir) do |path|
      if File.symlink?(path)
        lnk = File.expand_path(path)

        # Need to give it relative to lnk
        src = File.expand_path(File.readlink(path), File.dirname(lnk))

        # Only remove symlinks that are relative to dir. i.e. inside
        # project to other parts of the project.

        if src.index(dir) == 0

          symlinks[lnk] = src

          # Store if it is a directory
          dircheck = File.directory?(path)

          # Remove the link
          FileUtils.rm_f(path)

          if dircheck
            # Prune must come after removing the file, because code
            # after prune will not run.
            Find.prune
          end

        end
      end
    end

    yield dir
  ensure

    symlinks.each do |sym, source|
      begin
        FileUtils.ln_s(source, sym)
      rescue => err
        STDERR.puts "Cannot restore link:", err
      end
    end
  end

end
