# $Id: executable_files_check.rb 29 2008-04-02 07:32:44Z zev $
#
# This action checks for the code size within ruby scripts starting with
# a shell bang. By our convention, we want to keep those files as small
# as possible, all logic should be in reusable classes.

# Copyright (c) 2006, Ubiquitous Business Technology (http://ubit.com)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    * Neither the name of Ubit nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# == Author
# Michael Reinsch (mr@ubit.com)

require File.dirname(__FILE__) + "/loc_count.rb"

class ExecutableFilesCheckAction < BuildAction
  include LOCCountModule

  def initialize
    @problem_files = Array.new
    @shellbang_files_LOC = Hash.new
  end

  def build_action(context)
    context.ruby_files.each do |f|
      exec = File.executable?(f)
      bang = ("#!/" == IO.read(f, 3, 0))
      if exec ^ bang
        if exec
          prob = "executable, no shell bang"
        else
          prob = "shell bang, not executable"
        end
        @problem_files << { :filename => f, :problem => prob }
      end
      @shellbang_files_LOC[f] = count_loc(f) if bang
    end
    @sorted_shellbang_files_LOC = @shellbang_files_LOC.sort {|a,b| b[1] <=> a[1] }
  end

  def summary_display(context)
    template = TemplateFile.new(self.class.summary_template_file)

    res = {}
    if @problem_files.size > 0
      res[:problem_files] = { :files => @problem_files }
    end

    # get top 10
    top_ten_files = @sorted_shellbang_files_LOC[0..9]
    res[:files] = top_ten_files.map do |f, l|
      { :filename => f, :lines_of_code => l }
    end

    context.amrita_data[:executable_files_check_results] = res
    execfilecodesize_base = "#{context.project_name}_executable_files_check.html"
    context.amrita_data[:executable_files_check_details] =
      (Amrita::e(:a, :href => execfilecodesize_base) { "Executable Files Check Details" } )

    summary_expand(template, context)
  end

  def detailed_display(context)
    template = TemplateFile.new(self.class.detailed_template_file)

    res = {}
    if @problem_files.size > 0
      res[:problem_files] = { :files => @problem_files }
    end

    res[:files] = @sorted_shellbang_files_LOC.map do |f, l|
      { :filename => f, :lines_of_code => l }
    end

    context.amrita_data[:executable_files_check_results] = res

    det_res = ProjectBuilderUtils.expand_template(template, context.amrita_data)
    det_file = "#{context.output_directory}/#{context.project_name}_executable_files_check.html"
    [det_file, det_res]
  end

end
