# $Id: loc_count.rb 29 2008-04-02 07:32:44Z zev $
#
# This file implements the logic to count the lines of code in
# a source base.

# Copyright (c) 2006, Ubiquitous Business Technology (http://ubit.com)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    * Neither the name of Ubit nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# == Author
# Zev Blut (zb@ubit.com)

module LOCCountModule
  def count_loc(file)
    code = comments = blank = 0
    inside_embedded_doc = false

    IO.readlines(file).each do |line|
      if line =~ /\s*#/
        #ignore comments
        comments += 1
      elsif line =~ /^=begin/
        inside_embedded_doc = true
        comments += 1
      elsif line =~ /^=end/
        inside_embedded_doc = false
        comments += 1
      elsif inside_embedded_doc
        comments += 1
      elsif line =~ /^\s*$/
        #blank space to ignore
        blank += 1
      else
        code+= 1
      end
    end

    [code, comments, blank]
  end
end


class LocCountAction < BuildAction
  include LOCCountModule

  TOTAL = "TOTAL_LOC"
  PROD_CODE = "TOTAL_PRODUCTION_LOC"
  TEST_CODE = "TOTAL_TEST_LOC"

  def initialize
    @file_loc = Hash.new { |h, k| h[k] = Array.new }
  end

  def build_action(context)

    (context.ruby_files + context.test_files).each do |file|
      @file_loc[file] = count_loc(file)
    end

    @file_loc[TOTAL] = sum_count(@file_loc.keys)
    @file_loc[PROD_CODE] = sum_count(context.ruby_files)
    @file_loc[TEST_CODE] = sum_count(context.test_files)
  end

  def summary_display(context)
    template = TemplateFile.new(self.class.summary_template_file)

    ares = [ ["Total", @file_loc[TOTAL]],
    ["Production Code", @file_loc[PROD_CODE]],
    ["Test Code", @file_loc[TEST_CODE]] ].map do |ctype, cdata|
      code, comments, blank, files = *cdata
      ttl = code + comments + blank
      { :type => ctype, :total => ttl,
        :code => code, :code_per => percent(code, ttl),
        :comments => comments, :comments_per => percent(comments, ttl),
        :blank => blank, :blank_per => percent(blank, ttl) ,
        :files => files
      }
    end

    context.amrita_data[:loc_results] = { :entry => ares }
    loc_base = "#{context.project_name}_loc.html"
    context.amrita_data[:loc_details] =
     (Amrita::e(:a, :href => loc_base) { "Lines of Code Details" } )

    summary_expand(template, context)
  end

  def detailed_display(context)
    template = TemplateFile.new(self.class.detailed_template_file)

    ares = (@file_loc.keys - [TOTAL, PROD_CODE, TEST_CODE]).sort.map do |file|
      code, comments, blank = *(@file_loc[file])
      ttl = code + comments + blank
      { :file => file, :total => ttl,
        :code => code, :code_per => percent(code, ttl),
        :comments => comments, :comments_per => percent(comments, ttl),
        :blank => blank, :blank_per => percent(blank, ttl) }
    end

    context.amrita_data[:loc_details] = { :entry => ares }
    det_res = ProjectBuilderUtils.expand_template(template, context.amrita_data)
    det_file = "#{context.output_directory}/#{context.project_name}_loc.html"
    [det_file, det_res]
  end

  def percent(val, total)
    sprintf("%.2f%",  (val.to_f / total) * 100)
  end

  def sum_count(files)
    code = comments = blank = 0

    files.each do |file|
      fcode, fcomments, fblank = *( @file_loc[file] )
      code += fcode
      comments += fcomments
      blank += fblank
    end

    [code, comments, blank, files.size]
  end

end
