# $Id: unit_test.rb 33 2008-04-02 08:14:27Z zev $
#
# This file implements the logic to perform Unit Tests

# Copyright (c) 2006, Ubiquitous Business Technology (http://ubit.com)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    * Neither the name of Ubit nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# == Author
# Zev Blut, Pierre Baumard, Michael Reinsch, Paul McMahon

require 'kwala'

class UnitTestBuildAction < BuildAction

  RES_REGEXP = '(.*)Finished in (\d+\.\d+) seconds.(.*)\s*(\d+) tests?, ' +
   '(\d+) assertions?, (\d+) failures?, (\d+) errors?'

  RUBY = '/usr/bin/env ruby'

  def initialize
    @test_results = nil
    @results_by_version = nil
    @time_start = Time.now

    # The maximum number of violations to show in the summary
    @time_violations_limit = 5

    # time in seconds
    @time_error_threshold = 5.0
    @time_warning_threshold = 1.0
  end

  def build_action(context)
    time_start = Time.now
    @test_results = run_test_cases(context.test_files,
                                   context.pre_requires,
                                   context.output_directory)

    @filtered_entries = filter_test_results_for_summary( @test_results )

    notify_test_results
    @task_time = (Time.now - time_start).to_i.to_s + "s"
  end

  def summary_display(context)
    template = TemplateFile.new(self.class.summary_template_file)
    unit_base = "#{context.project_name}_unit_test.html"

    context.amrita_data[:unit_test_results] = {
      :unit_tests_link => Amrita::e(:a, :href => unit_base) { "Unit Test Details" },
      :ruby_version => ruby_version,
      :entry => @filtered_entries,
    }

    context.amrita_data[:process_run_time] = {
      :task_time => @task_time
    }

    timing_base = "#{context.project_name}_timing.html"
    context.amrita_data[:unit_test_timing] = {
      :timing_link => Amrita::e(:a, :href => timing_base) { "Timing Details" },
      :time_violations_limit => @time_violations_limit,
      :error_threshold => @time_error_threshold,
      :warning_threshold => @time_warning_threshold,
      :timing_stats =>  timing_stats( @test_results ),
      :time_violations => time_violations( @test_results )
    }

    if context.amrita_data[:unit_test_results][:entry]
      context.amrita_data[:unit_test_results][:entry].each do |entry|
        entry[:file_name_link] = Amrita::a(:href =>
         "#{unit_base}\##{entry[:file_name]}")  { entry[:file_name] }
      end
    end

    if context.amrita_data[:unit_test_timing][:time_violations]
      context.amrita_data[:unit_test_timing][:time_violations].each do |entry|
        entry[:file_name_link] = Amrita::a(:href =>
         "#{unit_base}\##{entry[:file_name]}")  { entry[:file_name] }
      end
    end

    summary_expand(template, context)
  end

  def detailed_display(context)
    context.amrita_data[:unit_test_results] = {
      :entry => @test_results
    }
    det_res = expand_unit_test_template(context.amrita_data[:unit_test_results])
    det_file = "#{context.output_directory}/#{context.project_name}_unit_test.html"
    [det_file, det_res]
  end

  # Should be a bit more complicated, but for now we punish any test failures.
  def score
    if @test_results.empty?
      10
    else
      1
    end
  end

  ####################
  # Protected Methods
  ####################
  protected

  def notify_test_results
    # only mail errors and failures
    mail_entries = @filtered_entries.find_all do |e|
      [:errors, :failures].find do |type|
        e.key?(type) && !ignorable?(e[type])
      end
    end

    if mail_entries.size > 0
      data = { :filtered_results => mail_entries, :test_results => @test_results }
      notify(:failure, data)
    else
      notify(:success)
    end
  end

  def ruby_version
    return @version if defined? @version
    @version  = `#{RUBY} -v`
  end

  def expand_unit_test_template(data)
    template = TemplateFile.new(self.class.detailed_template_file)
    data[:entry].each do |entry|
      entry[:file_name_ref] = Amrita::a(:name => entry[:file_name]) { entry[:file_name] }
      entry[:sys_errors] = entry[:sys_errors].empty? ? nil : entry[:sys_errors].to_a.map {|d| {:error => d} }
      entry[:warnings] = entry[:warnings].empty? ? nil : entry[:warnings].to_a.map {|d| {:warning => d} }
    end
    ProjectBuilderUtils.expand_template(template, data)
  end

  def timing_stats(results)
    results_by_version(results).collect do |version_name,results|
      {:version_name => version_name,
       :total_time => results.inject(0) { |sum,r| sum + r[:time].to_f},
       :errors => results.find_all {|x| x[:time].to_f >= @time_error_threshold}.size,
       :warnings => results.find_all {|x| x[:time].to_f >= @time_warning_threshold}.size,
      }
    end
  end

  def results_by_version(results)
    return @results_by_version unless @results_by_version.nil?

    @results_by_version = {}
    results.each do |result|
      result[:version].each do |version|
        name = version[:version_name]
        @results_by_version[name] = [] if @results_by_version[name].nil?
        @results_by_version[name] << version
      end
    end
    @results_by_version
  end

  def time_violations(results)
    violations = []
    results_by_version(results).each do |version_name,results|
      version_violations = results.sort {|x,y| -(x[:time].to_f <=> y[:time].to_f)}
      violations << version_violations[0, @time_violations_limit]
    end

    violations.flatten.collect do |x|
      {:time => template_timing_error_num(x[:time]),
       :file_name => x[:file_name],
       :version_name => x[:version_name]}
    end
  end

  def run_test_cases(test_files, pre_requires=nil, out_dir="./")
    # Run test cases collect warnings merged for all versions
    results = Array.new
    test_files.each do |f|
      entry = Hash.new
      stdout_data, stderr_data = execute_test_file(pre_requires, f)
      entry[:file_name] = f
      entry[:version] = parse_test_results(stdout_data.join("\n"), f)
      warnings, errors = stderr_data.partition {|r| r =~ /: warning:/}
      entry[:warnings] = warnings
      entry[:sys_errors] = errors
      results << entry
    end
    results
  end

  def execute_test_file(pre_requires, f)
    # XXX UGLY WEDGE FOR RAILS
    rails_mig = `RAILS_ENV=test rake cerego:db:empty 2>&1`
    rails_exitstatus = $?.exitstatus

    stdout_data = stderr_data = nil
    Open3.popen3(ruby_command_line(pre_requires, f)) do |stdin, stdout, stderr|
      stdout_data = time_out_readlines(stdout)
      stderr_data = time_out_readlines(stderr)
    end
    if rails_exitstatus != 0
      stderr_data = stderr_data.to_s + "\n RAILS MIG FAILURE \n" + rails_mig
    end

    [stdout_data, stderr_data]
  end

  def ruby_command_line(pre_requires, file)
    req_cmd = ""
    if pre_requires
      req_cmd = "-r #{pre_requires}"
    end
    "#{RUBY} #{req_cmd} #{file}"
  end

  # returns array of results for each version
  def parse_test_results(data, file)
    reg = /#{RES_REGEXP}/im
    if m = reg.match(data)
      results = Hash.new
      results[:file_name] = file
      results[:version_name] = ruby_version()
      results[:time] = m[2]
      results[:tests] = m[4]
      results[:assertions] = m[5]
      results[:failures] =  template_error_num(m[6])
      results[:errors] = template_error_num(m[7])
      results[:test_results] = parse_out_test_info(m[3]).map {|i| {:test_info => i}}
    else
      results = no_match_result(data, file)
    end
    [results]
  end

  # e.g. time-out
  def no_match_result(data, file)
    {
      :file_name => file,
      :version_name => ruby_version,
      :time => '',
      :tests => '',
      :assertions => '',
      :failures => '',
      :errors => a(:bgcolor, 'red') {"unexpected"},
      :test_results => [{:test_info => data}]
    }
  end

  def template_error_num(num_str)
    num_str.to_i > 0 ? a(:bgcolor => "red") { num_str } : num_str
  end

  def template_timing_error_num(num_str)
    if num_str.to_f >= @time_error_threshold
      a(:bgcolor => "red") { num_str }
    elsif num_str.to_f >= @time_warning_threshold
      a(:bgcolor => "yellow") { num_str }
    else
      num_str
    end
  end

  def filter_test_results_for_summary(orig_results)
    filtered = Array.new
    orig_results.each do |orig_result|
      file_errors = nil
      # Let's show test cases that have warnings or errors
      # in the Ruby program.
      if (!orig_result[:warnings].to_s.empty?) ||
       (!orig_result[:sys_errors].to_s.empty?)
        filtered.concat(orig_result[:version])
        next
      end
      orig_result[:version].each do |version|
        if version.key?(:errors) and !ignorable?(version[:errors])
          filtered << version
          next
        end
        if version.key?(:failures) and  !ignorable?(version[:failures])
          filtered << version
        end
      end
    end
    filtered
  end

  def ignorable?(data)
   (data.is_a?(String) && data.to_i == 0)
  end

  def parse_out_test_info(test_info)
    groups = group_test_info(test_info)

    results = Array.new
    groups.each do |line|
      line.scan(/(\d+)\) (Failure|Error):\s*(.*)/im) do |num, type, msg|
        results << "#{num}) #{type}\n#{msg}"
      end
    end
    results
  end

  def group_test_info(test_info)
    in_group = false
    groups = Array.new
    cur_group = Array.new

    test_info.split(/\n{2}/).each do |line|
      if !line.empty? && !in_group
        in_group = true
        if !cur_group.empty?
          groups << cur_group
        end
        cur_group = Array.new
        cur_group << line
      elsif line.empty?
        in_group = false
        if !cur_group.empty?
          groups << cur_group.join("\n")
          cur_group = Array.new
        end
      else
        cur_group << line
      end

    end

    if !cur_group.empty?
      groups << cur_group.join("\n")
    end

    groups
  end

end
