#! /usr/bin/ruby1.8

# Copyright (c) 2006, Ubiquitous Business Technology (http://ubit.com)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    * Neither the name of Ubit nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# == Author
# Zev Blut (zb@ubit.com)

puts "Starting #{Time.now}"

require 'getoptlong'
require 'kwala'

def usage
  puts "Usage: #{$0} [-d -p -a -y] -o"
  puts "       --project_name, -p (project name): The name of the project"
  puts "       --directory, -d (project base directory): The base directory of the project."
  puts "                       if not set then current directory is assumed."
  puts "       --action, -a (action name): The name of an action to execute."
  puts "                       if no action is set then all actions are assumed."
  puts "       --yaml, -y (yaml debug file name): The name of a file to save yaml debug information to."
  puts "       --output_dir, -o (output directory): The directory to output results to."
  puts "       --require,-r (file to require): A file to require for use when testing unit tests."
  puts "       --ri_cmp_1,-1 (old ri directory): The old ri output directory to compare against."
  puts "       --ri_cmp_2,-2 (new ri directory): The new ri output directory to compare against."
  puts "       --list_actions, -l: Lists the action names available."
  puts "       --notifications, -n: Activates notifications for action results."
  puts "       --vcs_num, -v: A string with the vcs version numbers to check"
  exit 1
end

def list_actions
  puts "Available actions are: "
  BuildAction.command_line_action_names.each do |name|
    puts " #{ name }"
  end
end

ProjectBuilderUtils.load_actions

actions = Array.new
context = BuildContext.new

context.project_directory = Dir.pwd
context.project_name = ""
debug_results = nil
begin
  options = GetoptLong.new(
    ["-p", "--project_name", GetoptLong::REQUIRED_ARGUMENT],
    ["-d", "--dir", GetoptLong::REQUIRED_ARGUMENT],
    ["-a", "--action", GetoptLong::REQUIRED_ARGUMENT],
    ["-y", "--yaml", GetoptLong::REQUIRED_ARGUMENT],
    ["-r", "--require", GetoptLong::REQUIRED_ARGUMENT],
    ["-1", "--ri_cmp_1", GetoptLong::REQUIRED_ARGUMENT],
    ["-2", "--ri_cmp_2", GetoptLong::REQUIRED_ARGUMENT],
    ["-e", "--extensions", GetoptLong::REQUIRED_ARGUMENT],
    ["-o", "--output_dir", GetoptLong::REQUIRED_ARGUMENT],
    ["-l", "--list_actions", GetoptLong::NO_ARGUMENT],
    ["-n", "--notifications", GetoptLong::NO_ARGUMENT],
    ["-v", "--vcs_num", GetoptLong::REQUIRED_ARGUMENT] )

  options.each do |opt, arg|
    case opt
    when "-p"
      context.project_name = arg
    when "-a"
      actions << arg
    when "-d"
      context.project_directory = arg
    when "-y"
      debug_results = arg
    when "-o"
      context.output_directory = arg
    when "-r"
      context.pre_requires = arg
    when "-1"
      context.ri_dir_1 = ProjectBuilderUtils.clean_trailing_slash(arg)
    when "-2"
      context.ri_dir_2 = ProjectBuilderUtils.clean_trailing_slash(arg)
    when "-e"
      require arg
      # required file should define get extensions
      extensions = get_extensions
    when "-l"
      list_actions
      exit 0
    when "-n"
      Notification.context = context
      Notification.notifications
      Notification.add_notification(EmailNotification.new)
      Notification.add_notification(JabberNotification.new)
    when "-v"
      context.vcs_num = arg
    else
      STDERR.puts "Unknown argument #{opt}"
      usage
    end
  end
rescue => err
  STDERR.puts "Error: #{err}", err.backtrace
  usage
end

usage if context.output_directory.nil?

# Set default actions if not passed
if actions.empty?
  actions = %w(loc_count syntax_check unit_test flagged_comments)
end

if actions.include?("code_change")
  usage if (context.ri_dir_1 && !context.ri_dir_2) || (!context.ri_dir_1 && context.ri_dir_2)
end

# Set up necessary info before actions

ruby_files, test_files = ProjectBuilderUtils.get_ruby_and_unit_test_files(context.project_directory)
context.ruby_files = ruby_files
context.test_files = test_files

action_instances = Array.new
actions.each do |action|
  action_obj = BuildAction.create_action_from_command_line_name(action)
  if action_obj
    action_instances << action_obj
  else
    STDERR.puts "Unknown action : #{action}"
    list_actions
  end
end

# Run the build stage
action_instances = ProjectBuilderUtils.build_actions(action_instances, context)

# Run the summary stage
summaries = ProjectBuilderUtils.build_summaries(action_instances, context)

if debug_results
  #Check the syncro of templates
  File.open(debug_results, "w") do |f|
    f.puts context.amrita_data.to_yaml
  end
end

ProjectBuilderUtils.make_detailed_reports(context, action_instances)

ProjectBuilderUtils.make_summary_report(context, summaries)


# Copy the stylesheet to the output directory
FileUtils.cp(TEMPLATE_DIR + "style.css", context.output_directory)
FileUtils.cp_r(TEMPLATE_DIR + "mdd", context.output_directory)

puts "Done #{Time.now}"
