require 'pstore'
require 'rss/2.0'
require 'rss/utils'

class RSSNotification < Notification

  def unit_test(result, data, context)
    @test_results = data[:test_results]

    update_pstore(context)
    make_rss(context)
  end

  protected

  PSTORE_FILE_NAME = 'prev_test_results.pstore'

  RSS_FILE_NAME = 'unit_test_results.rss'

  class UnitTestResult
    attr_reader :filename, :test_results, :errors, :warnings, :first_seen
    attr_accessor :last_seen

    def initialize(result_hash, timestamp)
      @filename = result_hash[:file_name]
      @errors = result_hash[:sys_errors]
      @warnings = result_hash[:warnings]
      @first_seen = @last_seen = timestamp

      @test_results = Hash.new
      result_hash[:version].each do |r|
        if !r[:test_results].empty?
          @test_results[r[:version_name]] = r[:test_results]
        end
      end
    end

    def pass?
      @errors.empty? && @warnings.empty? && @test_results.empty?
    end

    def ==(obj)
      res = obj.kind_of?(UnitTestResult)
      res &= obj.filename == @filename
      res &= obj.errors == @errors
      res &= obj.warnings == @warnings
      res &= obj.test_results == @test_results
      res
    end

    def to_rss_item(compare_timestamp)
      if @last_seen < compare_timestamp
        defect_status = "Fixed"
      else
        defect_status = "New"
      end

      item = RSS::Rss::Channel::Item.new
      item.title = "#{@first_seen.strftime("%b %d, %H:%M")} - #{defect_status} defects in #{@filename}"
      item.pubDate = @first_seen
      item.description = defect_details_html
      item
    end

    protected

    def system_messages_html(messages, type)
      details = ""
      if !messages.empty?
        details << "<h2>System #{type}s</h2>\n<pre>\n"
        details << RSS::Utils.html_escape(messages.join)
        details << "\n</pre>\n"
      end
      details
    end

    def test_result_html
      details = ""
      @test_results.each do |ruby_ver, defects|
        details << "<h2>Test Results for #{ruby_ver}</h2>\n"
        defects.each do |d|
          details << "<pre>\n"
          details << RSS::Utils.html_escape(d[:test_info])
          details << "\n</pre>\n"
        end
      end
      details
    end

    def defect_details_html
      details = ""
      details << test_result_html
      details << system_messages_html(@errors, "Error")
      details << system_messages_html(@warnings, "Warning")
      details
    end
  end

  def create_pstore(context)
    PStore.new(File.join(context.output_directory, PSTORE_FILE_NAME))
  end

  def update_pstore(context)
    create_pstore(context).transaction do |store|
      run_timestamp = Time.now
      del_timestamp = run_timestamp - 12*60*60 # 12 hours
      prev_test_results = store.fetch(:prev_test_results, Hash.new)

      prev_test_results.delete_if do |key, res|
        res.last_seen <= del_timestamp
      end

      @test_results.each do |cur_res_hash|
        cur_res = UnitTestResult.new(cur_res_hash, run_timestamp)
        next if cur_res.pass?

        prev_res = prev_test_results[cur_res.filename]
        if prev_res.nil? || cur_res != prev_res
          # new defects
          prev_test_results[cur_res.filename] = cur_res
        else
          # same'ol defect
          prev_res.last_seen = run_timestamp
        end
      end

      store[:last_run_timestamp] = run_timestamp
      store[:prev_test_results] = prev_test_results
    end
  end

  def make_rss(context)
    rss = RSS::Rss.new("2.0")
    chan = RSS::Rss::Channel.new
    chan.title = chan.description = "Unit Test Results for #{context.project_name}"
    rss.channel = chan

    create_pstore(context).transaction(true) do |store|
      last_run_timestamp = store.fetch(:last_run_timestamp, 0)
      test_results = store.fetch(:prev_test_results, Hash.new)

      test_results.each do |key, res|
        chan.items << res.to_rss_item(last_run_timestamp)
      end
    end

    File.open(File.join(context.output_directory, RSS_FILE_NAME), "w") {|f| f.puts(rss.to_s) }
  end


end
