module ProjectBuilderUtils

  def ProjectBuilderUtils.find_ruby_files(dir)
    ruby_files = Array.new

    Find.find(dir) do |path|
      if File.directory?(path)
        # Ignore all directories starting with .
        if File.basename(path)[0] == ?. || File.basename(path) =~ /vendor/
          Find.prune
        else
          next
        end
      elsif File.file?(path)
        case path.to_s
        when /\.rb$/
          ruby_files << path
        when /\.rb~$/
          next
        when ProjectBuilderUtils.contains_ruby_execution_command?(path)
          ruby_files << path
        else
        end
      end
    end

    ruby_files = ruby_files.sort
  end

  def ProjectBuilderUtils.strip_base_dir(base_dir, file)
    m = /(#{base_dir})(.*)/.match(file)
    m[2][1..-1]
  end

  def ProjectBuilderUtils.get_ruby_and_unit_test_files(dir=Dir.pwd)
    ruby_files = ProjectBuilderUtils.find_ruby_files(dir)

    # remove all test case files
    test_files, ruby_files = ruby_files.partition do |f|
      f =~ /\/(tc|ts)_.+\.rb$|.*_test.rb$/
    end

    # XXX: UGLY HACK TO AVOID SIMULATION
    test_files.delete_if { |f| f =~ /test\/simulation\// }

    [ruby_files.uniq, test_files.sort.uniq]
  end

  def ProjectBuilderUtils.clean_trailing_slash(dir)
    if dir[dir.size - 1] == ?/
      if dir.size > 1
        dir = dir[0..dir.size - 2]
      else
        dir = "."
      end
    end
    dir
  end


  def ProjectBuilderUtils.expand_template(template, data)
    str = StringIO.new
    template.expand(str, data)
    str.string
  end

  def ProjectBuilderUtils.load_actions
    Dir.glob(ACTION_DIR + "/*.rb").sort.each do |action_file|
      require(action_file)
    end
  end

  def ProjectBuilderUtils.contains_ruby_execution_command?(path)
    File.open(path, "r") do |f|
      return (f.read(1024).to_s =~ /^#!.*ruby/)
    end
  end


  def ProjectBuilderUtils.expand_summary_template(data, project_name, summaries)
    template = TemplateFile.new("#{TEMPLATE_DIR}/build_template.html")
    data[:action_summaries] = summaries
    ProjectBuilderUtils.expand_template(template, data)
  end

  def ProjectBuilderUtils.make_summary_report(context, summaries)
    if context.vcs_num

      context.amrita_data[:vcs] = {
        :vcs_num => context.vcs_num,
        :vcs_info => context.vcs_info
      }
    end
    context.build_complete

    exp_res = ProjectBuilderUtils.expand_summary_template(context.amrita_data, context.project_name, summaries)
    if context.output_directory
      sum_file = "#{context.output_directory}/index.html"
      File.open(sum_file, "w") do |f|
        f.puts exp_res
      end
    else
      STDOUT.puts exp_res
    end
  end

  # Pull this into better place
  def ProjectBuilderUtils.calc_score(context, action_instances)
    scores = action_instances.map { |a| a.score }
    scores.compact!

    # If action doesn't calculate score, just default to 1 for now
    if !scores.empty?
      total = scores.inject(0) { |s,v| s + v }

      #    puts "Total : #{total} SIZE: #{scores.size}"
      score = ((total / (scores.size * 10).to_f) * 10).ceil
      img_num = ProjectBuilderUtils.image_for_score(score)
      #    puts "!!!!!!!!!!!!!!!!!! SCORE : #{score} IMG #{img_num} !!!!!!!!!!!"
    else
      img_num = 1
    end
    # Mascot Driven Development!
    context.amrita_data[:mdd] = Amrita.a(:src => "mdd/#{img_num}.png")
  end

  def ProjectBuilderUtils.image_for_score(score)
    case score
    when 8..10 : 1
    when 6..7 : 2
    when 4..5 : 3
    when 3..2 : 4
    else
      5
    end
  end

  def ProjectBuilderUtils.make_detailed_reports(context, action_instances)
    if context.output_directory
      action_instances.each do |action|
        det_file, det_res = action.detailed_display(context)
        if det_file
          File.open(det_file, "w") do |f|
            f.puts det_res
          end
        end
      end
    end

    ProjectBuilderUtils.calc_score(context, action_instances)
  end

  # returns an array of the actions that were successfully built
  def ProjectBuilderUtils.build_actions(action_instances, context)
    success = Array.new

    action_instances.each do |action|  
      begin
        action.build_action(context)
        success << action
      rescue KwalaAbortBuildException => err        
        # We want to failed action, but stop the others
        success << action
        break
      end
    end    

    success
  end

  def ProjectBuilderUtils.build_summaries(action_instances, context)
    action_instances.map do |action|
      { :summary => action.summary_display(context) }
    end
  end


end
